use crate::memory::address::{PhysicalPageNumber, VirtualPageNumber, PHYSICAL_PAGE_NUMBER_WIDTH};
use crate::memory::frame_allocator::{allocate_frame, FrameTracker};
use alloc::vec;
use alloc::vec::Vec;
use bitflags::bitflags;

bitflags! {
    #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
    pub struct PageTableEntryFlags: u8 {
        /// Valid
        const V = 1 << 0;
        ///Read
        const R = 1 << 1;
        /// Write
        const W = 1 << 2;
        /// Execute
        const X = 1 << 3;
        /// User
        const U = 1 << 4;
        /// unknown
        const G = 1 << 5;
        /// Accessed
        const A = 1 << 6;
        /// Dirty
        const D = 1 << 7;
    }
}

#[derive(Copy, Clone)]
#[repr(C)]
pub struct PageTableEntry {
    pub bits: usize,
}

impl PageTableEntry {
    pub fn new(page_num: PhysicalPageNumber, flags: PageTableEntryFlags) -> Self {
        Self {
            bits: page_num.0 << 10 | flags.bits() as usize,
        }
    }

    pub fn empty() -> Self {
        Self { bits: 0 }
    }

    pub fn physical_page_numer(&self) -> PhysicalPageNumber {
        (self.bits >> 10 & (1usize << PHYSICAL_PAGE_NUMBER_WIDTH - 1)).into()
    }

    pub fn flags(&self) -> PageTableEntryFlags {
        PageTableEntryFlags::from_bits(self.bits as u8).unwrap()
    }

    pub fn is_valid(&self) -> bool {
        (self.flags() & PageTableEntryFlags::V) != PageTableEntryFlags::empty()
    }

    pub fn readable(&self) -> bool {
        (self.flags() & PageTableEntryFlags::R) != PageTableEntryFlags::empty()
    }

    pub fn writable(&self) -> bool {
        (self.flags() & PageTableEntryFlags::W) != PageTableEntryFlags::empty()
    }

    pub fn executable(&self) -> bool {
        (self.flags() & PageTableEntryFlags::X) != PageTableEntryFlags::empty()
    }
}

pub struct PageTable {
    root_page_number: PhysicalPageNumber,
    frames: Vec<FrameTracker>,
}

impl PageTable {
    pub fn new() -> Self {
        let frame = allocate_frame().unwrap();
        Self {
            root_page_number: frame.page_number,
            frames: vec![frame],
        }
    }

    pub fn map(
        &mut self,
        virtual_page_number: VirtualPageNumber,
        physical_page_number: PhysicalPageNumber,
        flags: PageTableEntryFlags,
    ) {
        let entry = self.find_entry_or_create(virtual_page_number);
        assert!(
            !entry.is_valid(),
            "Virtual page number {:?} is mapped before mapping.",
            virtual_page_number
        );
        *entry = PageTableEntry::new(physical_page_number, flags | PageTableEntryFlags::V);
    }

    pub fn unmap(&mut self, virtual_page_number: VirtualPageNumber) {
        let entry = self.find_entry(virtual_page_number).unwrap();
        assert!(
            entry.is_valid(),
            "Virtual page number {:?} is invalid before unmapping",
            virtual_page_number
        );
        *entry = PageTableEntry::empty();
    }

    /// 从satp寄存器创建临时的页表
    /// 方便在用户空间时手动查看页表
    pub fn from_token(satp: usize) -> Self {
        Self {
            root_page_number: PhysicalPageNumber::from(
                satp & ((1usize << PHYSICAL_PAGE_NUMBER_WIDTH) - 1),
            ),
            frames: vec![],
        }
    }

    pub fn translate(&self, virtual_page_number: VirtualPageNumber) -> Option<PageTableEntry> {
        self.find_entry(virtual_page_number).map(|e| e.clone())
    }

    fn find_entry_or_create(&mut self, page_number: VirtualPageNumber) -> &mut PageTableEntry {
        let indexes = page_number.get_indexes();
        let mut physical_page_number = self.root_page_number;

        for i in 0..2 {
            let entry = &mut physical_page_number.get_page_table_entry_array()[indexes[i]];

            // 当V是1，其他都是0的页帧就是页表
            if !entry.is_valid() {
                let frame = allocate_frame().unwrap();
                *entry = PageTableEntry::new(frame.page_number, PageTableEntryFlags::V);
                self.frames.push(frame);
            }

            physical_page_number = entry.physical_page_numer();
        }

        &mut physical_page_number.get_page_table_entry_array()[indexes[2]]
    }

    fn find_entry(&self, page_number: VirtualPageNumber) -> Option<&mut PageTableEntry> {
        let indexes = page_number.get_indexes();
        let mut physical_page_number = self.root_page_number;

        for i in 0..2 {
            let entry = &mut physical_page_number.get_page_table_entry_array()[indexes[i]];

            if !entry.is_valid() {
                return None;
            }

            physical_page_number = entry.physical_page_numer();
        }

        Some(&mut physical_page_number.get_page_table_entry_array()[indexes[2]])
    }
}
